"""
Test the sensitivity of nest results on input parameters
"""

import numpy as np
import scanpy as sc
import pandas as pd
import os
import matplotlib.pyplot as plt
import seaborn as sns
from sklearn.metrics import jaccard_score, normalized_mutual_info_score
from scipy.optimize import linear_sum_assignment

import random
from collections import defaultdict

import nest
from nest.synthetic.data import create_synthetic_data 
from anndata import AnnData
from typing import Optional, Tuple
from numpy.typing import ArrayLike



def score_similarity(adata):
    layer_matrix = adata.uns['layer_matrix']
    ch_df = sc.get.obs_df(adata, keys=[v for v in adata.obs.columns if "hotspots_multi" in v])
    multi_hotspots = np.array(pd.notnull(ch_df)).astype(np.int_)
    n_pixels = layer_matrix.shape[0]
    
    def compute_sim_vals(arr, ref_denom=None):
        arr = arr.copy().astype(np.float_)
        arr /= (np.sum(arr, axis=1, keepdims=True)+1e-8)
        a1 = arr[:, np.newaxis, :]
        a2 = arr[np.newaxis, :, :]
        v = np.sum(a1*a2, axis=-1)

        d1 = np.diag(v).reshape(-1, 1)
        d2 = np.diag(v).reshape(1, -1)

        denom = (d1+d2+1e-20)
        if ref_denom is not None:
            denom = np.maximum(denom, ref_denom)

        sim = (2*v)/denom

        return sim, denom

    s1, d = compute_sim_vals(layer_matrix)
    s2, _ = compute_sim_vals(multi_hotspots, d)
    
    score = 1 - (np.linalg.norm(s1-s2)**2)/(n_pixels**2)

    return score

def pairwise_jaccard(adata, label=None):
    layer_matrix = adata.uns['layer_matrix']
    if label is None:
        ch_df = sc.get.obs_df(adata, keys=[v for v in adata.obs.columns if "hotspots_multi" in v])
        multi_hotspots = np.array(pd.notnull(ch_df)).astype(np.int_)
    else:
        multi_hotspots = pd.get_dummies(adata.obs[label]).to_numpy().astype(np.int_)
    n_pixels = layer_matrix.shape[0]

    sim_matrix = np.zeros(shape=[layer_matrix.shape[1], multi_hotspots.shape[1]])

    for idx in range(layer_matrix.shape[1]):
        labels_1 = layer_matrix[:, idx]
        for idy in range(multi_hotspots.shape[1]):
            labels_2 = multi_hotspots[:, idy]
            sim_matrix[idx, idy] = jaccard_score(labels_1, labels_2)

    row_ind, col_ind = linear_sum_assignment(1-sim_matrix)

    res = np.zeros(max(layer_matrix.shape[1], multi_hotspots.shape[1]))
    res[:len(row_ind)] = sim_matrix[row_ind, col_ind]
       
    return np.mean(res)  


def seg_similarity(adata, label):
    regions = adata.obs['regions']
    score = normalized_mutual_info_score(regions, adata.obs[label])
    return score



if __name__=="__main__":
    # Tuning parameters
    # First value is default/base, others are variations to test for sensitivity
    # Test varying one parameter at a time (leave others at base)
    neighbor_eps_v = [0.1, 0.02, 0.05, 0.2]
    density_v = [0.2, 0.05, 0.1, 0.4, 0.6, 0.8]
    jaccard_threshold_v = [0.3, 0.15, 0.2, 0.25, 0.35, 0.4, 0.45]
    #neighbor_eps_v = [0.1]
    #density_v = [0.2]
    #jaccard_threshold_v = [0.3]
    #resolution_v = [1.0, 0.0, 0.1, 0.5, 2.0, 5.0, 10.0]
    resolution_v = np.linspace(0.0, 10.0, 20)
    params_list = [neighbor_eps_v, 
                   density_v, jaccard_threshold_v, resolution_v]


    # Non-tuning parameters
    hotspot_min_size = 20
    min_genes = 3
    dataset = "synthetic_hierarchy"

    # Configuration for plotting and saving
    dataset_dir = os.path.expanduser("~/Documents/data/")
    cache_dir = os.path.expanduser(f"data/{dataset}")
    image_save_dir = os.path.expanduser(f"images/{dataset}/")
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    if not os.path.exists(image_save_dir):
        os.makedirs(image_save_dir)

    np.random.seed(5)

    columns = defaultdict(list)
    for spatial_genes in [16, 32, 64, 128, 256, 512]:
        for trial in range(1):
            adata = create_synthetic_data(n_pixels = 64, n_genes = 2048, spatial_genes = spatial_genes, n_layers = 5, log1p=True, type="hierarchy",
                                            dropout_active = 0.5,
                                            mean_low=1.0, mean_high=5.0)

            adata_base = adata.copy()
            adata_single_gene_hotspots = adata.copy()
            min_samples = nest.get_min_samples_from_density(adata, neighbor_eps_v[0], density_v[0])
            res = nest.compute_gene_hotspots(adata_single_gene_hotspots, verbose=True, log=True,
                                                    eps=neighbor_eps_v[0], min_samples=min_samples, min_size=hotspot_min_size)
            adata_ref = adata_single_gene_hotspots.copy()
            #nest.coexpression_hotspots(adata_ref, threshold=0.3, min_size=1, cutoff=0.5, min_genes=3, resolution=1.000)
            nest.coexpression_hotspots(adata_ref, threshold=jaccard_threshold_v[0], min_size=1, cutoff=0.3, min_genes=min_genes,
                                        resolution=resolution_v[0], processes=4)
            
            
            nest.compute_multi_boundaries(adata_ref, 0.005, 0.00001, verbose=False)
            nest.multi_closure(adata_ref)

            score = pairwise_jaccard(adata_ref)

            print(score)

            columns['spatial genes'].append(spatial_genes)
            columns['nest score'].append(score)

            hmrf = nest.hmrf.HMRFSegmentationModel(adata, regions=5, k=4, label_name="class_hmrf")
            hmrf.fit(max_iterations=200, verbose=True, update_labels=True)

            jaccard = pairwise_jaccard(adata, label="class_hmrf")
            nmi = seg_similarity(adata, "class_hmrf")
            print(nmi)
            print(jaccard)

            columns['hmrf nmi'].append(nmi)
            columns['hmrf jaccard'].append(jaccard)

            spagcn = nest.methods.SpaGCN(regions=5)
            spagcn.fit(adata, verbose=True)

            jaccard = pairwise_jaccard(adata, label="class_spagcn")
            nmi = seg_similarity(adata, "class_spagcn")

            print(nmi)
            print(jaccard)

            columns['spagcn nmi'].append(nmi)
            columns['spagcn jaccard'].append(jaccard)



    df = pd.DataFrame(data=columns)
    df.to_csv(f'data/synthetic_hierarchy/synthetic_seg.csv')


"""
Test hierarchical agglomerative clustering algorithms for comparison
"""

import numpy as np
import scanpy as sc
import pandas as pd
import squidpy as sq
import os
import matplotlib.pyplot as plt
import seaborn as sns

import random

import nest
from nest.synthetic.data import create_synthetic_data 
from anndata import AnnData
from typing import Optional, Tuple
from numpy.typing import ArrayLike

from sklearn.cluster import AgglomerativeClustering
from scipy.cluster.hierarchy import dendrogram, linkage

def get_groups(model):
    membership_dict = dict()
    n_samples = len(model.labels_)
    for i, merge in enumerate(model.children_):
        cur_list = []
        for child_idx in merge:
            if child_idx < n_samples:
                cur_list.append(child_idx)
            else:
                cur_list.extend(membership_dict[child_idx - n_samples])
        membership_dict[i] = cur_list

    return membership_dict

def plot_dendrogram(model, **kwargs):
    # Create linkage matrix and then plot the dendrogram

    # create the counts of samples under each node
    counts = np.zeros(model.children_.shape[0])
    n_samples = len(model.labels_)
    for i, merge in enumerate(model.children_):
        current_count = 0
        for child_idx in merge:
            if child_idx < n_samples:
                current_count += 1  # leaf node
            else:
                current_count += counts[child_idx - n_samples]
        counts[i] = current_count

    linkage_matrix = np.column_stack(
        [model.children_, model.distances_, counts]
    ).astype(float)

    # Plot the corresponding dendrogram
    dendrogram(linkage_matrix, **kwargs)

if __name__ == "__main__":
    dataset = "V1_Breast_Cancer_Block_A_Section_1"

    # Configuration for plotting and saving
    dataset_dir = os.path.expanduser("~/Documents/data/")
    cache_dir = os.path.expanduser(f"data/{dataset}")
    image_save_dir = os.path.expanduser(f"images/{dataset}/")
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    if not os.path.exists(image_save_dir):
        os.makedirs(image_save_dir)
    nest.plot.set_dataset_plot_parameters(dataset)

    #adata = sq.datasets.slideseqv2()
    #adata = create_synthetic_data(n_pixels = 64, n_genes = 256, spatial_genes = 256, n_layers = 10, log1p=True, type="hierarchy",
    #                                   dropout_active = 0.5,
    #                                   mean_low=1.0, mean_high=5.0)
    adata = nest.data.get_data(dataset, dataset_dir, normalize=True)
    n_pca = 15
    sc.pp.pca(adata, n_comps=n_pca)

    cluster = AgglomerativeClustering(distance_threshold=None, n_clusters=5, linkage="ward")
    cluster.fit(adata.obsm['X_pca'])

    membership = get_groups(cluster)

    for i in range(adata.shape[0]-2, adata.shape[0]-60, -1):
        tmp = membership[i]
        v = np.zeros(adata.shape[0], dtype=np.int_)
        v[tmp] = 1
        adata.obs['tmp'] = pd.Categorical(v, categories=[1])
        fig, ax = plt.subplots(1, 1, figsize=[1, 1])
        nest.plot.spatial(adata, color="tmp", show=False, ax=ax, title="", frameon=False, legend_loc=None, alpha_img=0.5)
        fig.savefig(os.path.expanduser(f"~/Desktop/images/agglomerative_{adata.shape[0] - 2 - i}.png"), dpi=300, transparent=True, bbox_inches="tight")
        plt.close(fig)


    #adata.obs['tmp'] = pd.Categorical(cluster.labels_)
    #nest.plot.spatial(adata, color="tmp", show=True)

    #plot_dendrogram(cluster, truncate_mode="level", p=3)
    #plt.show()


"""
Test the sensitivity of nest results on input parameters
"""

import numpy as np
import scanpy as sc
import pandas as pd
import os
import matplotlib.pyplot as plt

import random

import nest
from anndata import AnnData
from typing import Optional, Tuple
from numpy.typing import ArrayLike





if __name__=="__main__":
    # Tuning parameters
    # First value is default/base, others are variations to test for sensitivity
    # Test varying one parameter at a time (leave others at base)
    use_smoothing_v = [True, False]
    use_modified_otsu_scaling_v = [True, False]
    neighbor_eps_v = [75, 50, 100]
    # For sake of simplicity we scale this later on proportional to neighbor_eps**2, the area being considered
    min_samples_v = [5, 3, 4, 8, 10]
    jaccard_threshold_v = [0.35, 0.2, 0.25, 0.3, 0.4, 0.45]
    resolution_v = [1.0, 0.0, 0.5, 1.5, 2.0, 5.0]
    params_list = [use_smoothing_v, use_modified_otsu_scaling_v, neighbor_eps_v, 
                   min_samples_v, jaccard_threshold_v, resolution_v]


    # Non-tuning parameters
    hotspot_min_size = 50
    min_genes = 8
    jaccard_cutoff = 0.5


    dataset = "slideseq"

    # Configuration for plotting and saving
    dataset_dir = os.path.expanduser("~/Documents/data/")
    cache_dir = os.path.expanduser(f"data/{dataset}")
    image_save_dir = os.path.expanduser(f"images/{dataset}/")
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    if not os.path.exists(image_save_dir):
        os.makedirs(image_save_dir)
    nest.plot.set_dataset_plot_parameters(dataset)

    adata = nest.data.get_data(dataset, dataset_dir, normalize=True)

    # Precompute reference parameter versions, allowing for reuse later
    if 1:
        adata_unsmoothed = adata.copy()
        adata_smoothed = adata.copy()
        nest.spatial_smoothing(adata_smoothed)

        adata_single_gene_hotspots = adata_smoothed.copy()
        nest.compute_gene_hotspots(adata_single_gene_hotspots, verbose=True, log=use_modified_otsu_scaling_v[0],
                                                eps=neighbor_eps_v[0], min_samples=min_samples_v[0], min_size=hotspot_min_size)
        
        adata_ref = adata_single_gene_hotspots.copy()
        nest.coexpression_hotspots(adata_ref, threshold=jaccard_threshold_v[0], min_size=30, cutoff=jaccard_cutoff, min_genes=min_genes,
                                    resolution=resolution_v[0])
        
        
        nest.compute_multi_boundaries(adata, 0.005, 0.00001, verbose=False)

        fig, ax = plt.subplots()
        nest.plot.multi_hotspots(adata, ax=ax, show=False)

        fig.savefig(os.path.join(image_save_dir, f"multi_hotspots_ref.png"), dpi=300, bbox_inches='tight', transparent=True)



    c = 0
    param_to_vary = 0
    for param_to_vary in range(6):
        cur_param_idx = 0 # range(len(params_list[param_to_vary]))
        for cur_param_idx in range(1, len(params_list[param_to_vary])):
            cur_params_list = [params_list[k][cur_param_idx] if k == param_to_vary else params_list[k][0] for k in range(len(params_list))]

            use_smoothing, use_modified_otsu_scaling, neighbor_eps, min_samples, jaccard_threshold, jaccard_resolution = cur_params_list
            
            # rescale min_samples
            min_samples = np.ceil(min_samples * (neighbor_eps/neighbor_eps_v[0])**2).astype(np.int_)

            print(c, use_smoothing, use_modified_otsu_scaling, neighbor_eps, min_samples, jaccard_threshold, jaccard_resolution )
            c += 1

            if param_to_vary < 4:
                if use_smoothing:
                    adata = adata_smoothed.copy()
                else:
                    adata = adata_unsmoothed.copy()

                res = nest.compute_gene_hotspots(adata, verbose=False, log=use_modified_otsu_scaling,
                                                eps=neighbor_eps, min_samples=min_samples, min_size=hotspot_min_size)
            else:
                adata = adata_single_gene_hotspots.copy()

            
            jaccard_matrix, hotspot_lookup = nest.hotspot_jaccard(adata, processes=10,
                                                                verbose=False, divisions=10)
            
            nest.coexpression_hotspots(adata, threshold=jaccard_threshold, min_size=30, cutoff=jaccard_cutoff, min_genes=min_genes,
                                    resolution=jaccard_resolution)
            try:
                nest.compute_multi_boundaries(adata, 0.005, 0.00001, verbose=False)
            except ValueError:
                continue

            # Compute the similarity over each chosen hotspot

            fig, ax = plt.subplots()
            nest.plot.multi_hotspots(adata, ax=ax, show=False)

            fig.savefig(os.path.join(image_save_dir, f"multi_hotspots_{use_smoothing}_{use_modified_otsu_scaling}_{neighbor_eps}_{min_samples}_{jaccard_threshold}_{jaccard_resolution}.png"), dpi=300, bbox_inches='tight', transparent=True)



import nest
import os
import anndata
import scipy
import sklearn.metrics
import numpy as np
import pandas as pd
import seaborn as sns
import scanpy as sc
import squidpy as sq
import networkx as nx
from pathlib import Path
import time
from tqdm import tqdm
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib import rc
import networkx as nx
import colorcet as cc

import warnings
warnings.filterwarnings("ignore")


def load_dataset():
    dataset = "slideseq"
    neighbor_eps = 75
    min_samples = 5
    hotspot_min_size = 50

    adata = nest.data.get_data(dataset, normalize=True)

    # Note: we perform this spatial smoothing for this specific dataset
    nest.spatial_smoothing(adata)






if __name__=="__main__":
    # Load and process slideseq data
    pass


import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import os

import nest


def filter_param(df, param):
    filter_set = {1, 2, 3, 4} - {param}
    return df[~df.varied.isin(filter_set)]


if __name__ == "__main__":
    multi_dfs = []
    for spatial_genes in [24, 32, 64, 256, 512]:
        dfs = [pd.read_csv(f'data/synthetic_hierarchy/results_{spatial_genes}_{trial}.csv') for trial in range(1)]
        df = pd.concat(dfs)
        df['spatial_genes'] = spatial_genes
        multi_dfs.append(df)

    combined_df = pd.concat(multi_dfs)

    combined_df['spatial_genes'] = pd.Categorical(combined_df['spatial_genes'])

    sns.set(style="whitegrid", font_scale=0.75)
    
    fig, axs = plt.subplots(1, 4, figsize=[6, 2])

    sns.lineplot(filter_param(combined_df, 0), x="eps", y="score", hue="spatial_genes", ax=axs[0], legend=None)
    axs[0].set_xlim(left=0.05, right=0.2)
    axs[0].set_xlabel('ε')
    axs[0].set_ylabel('Jaccard similarity')

    sns.lineplot(filter_param(combined_df, 1), x="density", y="score", hue="spatial_genes", ax=axs[1], legend=None)
    #axs[0].set_xlim(left=0.05, right=0.2)
    axs[1].set_ylabel('')

    sns.lineplot(filter_param(combined_df, 2), x="threshold", y="score", hue="spatial_genes", ax=axs[2], legend=None)
    #axs[0].set_xlim(left=0.05, right=0.2)
    axs[2].set_ylabel('')

    sns.lineplot(filter_param(combined_df, 3), x="resolution", y="score", hue="spatial_genes", ax=axs[3])
    #axs[0].set_xlim(left=0.05, right=0.2)
    axs[3].set_ylabel('')
    axs[3].legend(title="spatial genes")
    sns.move_legend(axs[3], "center left", bbox_to_anchor=(1, 0.5))

    for k in range(4):
        if k != 0:
            axs[k].set_ylabel('')
            axs[k].axes.yaxis.set_ticklabels([])
        axs[k].set_ylim([0, 1])



    
    fig.savefig(os.path.expanduser('~/Desktop/synthetic_hierarchy.pdf'), dpi=300, transparent=True, bbox_inches="tight")
    plt.show()


"""
Test the sensitivity of nest results on input parameters
"""

import numpy as np
import scanpy as sc
import pandas as pd
import os
import matplotlib.pyplot as plt

import random
import seaborn as sns

import nest
from anndata import AnnData
from typing import Optional, Tuple
from numpy.typing import ArrayLike

from nest.synthetic.data import score_similarity_adata
from sklearn.metrics import jaccard_score


def compute_max_jaccard(adata, adata_comp):
    ch_df_true = np.array(pd.notnull(sc.get.obs_df(adata, keys=[v for v in adata.obs.columns if "hotspots_multi" in v]))).astype(np.int_)
    ch_df_comp = np.array(pd.notnull(sc.get.obs_df(adata_comp, keys=[v for v in adata_comp.obs.columns if "hotspots_multi" in v]))).astype(np.int_)

    jaccard_scores = []
    for idx in range(ch_df_true.shape[1]):
        max_jaccard = max([jaccard_score(ch_df_true[:, idx], ch_df_comp[:, idy]) for idy in range(ch_df_comp.shape[1])])
        jaccard_scores.append(max_jaccard)

    return jaccard_scores


if __name__=="__main__":
    # Tuning parameters
    # First value is default/base, others are variations to test for sensitivity
    # Test varying one parameter at a time (leave others at base)
    use_smoothing_v = [True, False]
    use_modified_otsu_scaling_v = [True, False]
    neighbor_eps_v = [75, 50, 100]
    # For sake of simplicity we scale this later on proportional to neighbor_eps**2, the area being considered
    min_samples_v = [5, 3, 4, 8, 10]
    jaccard_threshold_v = [0.35, 0.2, 0.25, 0.3, 0.4, 0.45]
    resolution_v = [1.0, 0.0, 0.5, 1.5, 2.0, 5.0]
    params_list = [use_smoothing_v, use_modified_otsu_scaling_v, neighbor_eps_v, 
                   min_samples_v, jaccard_threshold_v, resolution_v]


    # Non-tuning parameters
    hotspot_min_size = 50
    min_genes = 8
    jaccard_cutoff = 0.5


    dataset = "slideseq"

    # Configuration for plotting and saving
    dataset_dir = os.path.expanduser("~/Documents/data/")
    cache_dir = os.path.expanduser(f"data/{dataset}")
    image_save_dir = os.path.expanduser(f"images/{dataset}/")
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    if not os.path.exists(image_save_dir):
        os.makedirs(image_save_dir)
    nest.plot.set_dataset_plot_parameters(dataset)

    adata = nest.data.get_data(dataset, dataset_dir, normalize=True)

    adata_smoothed = adata.copy()
    nest.spatial_smoothing(adata_smoothed)

    trials = 5
    
    frac_genes = 0.8
    n_genes = int(np.round(frac_genes*adata.shape[1]))

    # Compute hotspots on original adata
    if 1:
        adata_full = adata_smoothed.copy()
        nest.compute_gene_hotspots(adata_full, verbose=True, log=use_modified_otsu_scaling_v[0],
                                                    eps=neighbor_eps_v[0], min_samples=min_samples_v[0], min_size=hotspot_min_size)
        
        adata_ref = adata_full.copy()
        nest.coexpression_hotspots(adata_full, threshold=jaccard_threshold_v[0], min_size=30, cutoff=jaccard_cutoff, min_genes=min_genes,
                                    resolution=resolution_v[0])
    
        nest.compute_multi_boundaries(adata_full, 0.005, 0.00001, verbose=False)

    res = dict()

    for trial in range(trials):
        # Subsample genes
        gene_list = random.sample(list(adata.var_names), k=n_genes)
        adata_cur = adata_smoothed[:, gene_list].copy()

        # Run analysis
        nest.compute_gene_hotspots(adata_cur, verbose=True, log=use_modified_otsu_scaling_v[0],
                                                eps=neighbor_eps_v[0], min_samples=min_samples_v[0], min_size=hotspot_min_size)
        
        nest.coexpression_hotspots(adata_cur, threshold=jaccard_threshold_v[0], min_size=30, cutoff=jaccard_cutoff, min_genes=min_genes,
                                    resolution=resolution_v[0])
        
        #nest.compute_multi_boundaries(adata_cur, 0.005, 0.00001, verbose=False)
        #fig, ax = plt.subplots()
        #nest.plot.multi_hotspots(adata_cur, ax=ax, show=False)

        # compute overlap between each pair of hotspots
        jaccard_scores = compute_max_jaccard(adata_full, adata_cur)
        print(jaccard_scores)

        res[f'trial_{trial}'] = jaccard_scores

        #plt.show()


        #print(score_similarity_adata(adata_full, adata_cur))
        
        #fig.savefig(os.path.join(image_save_dir, f"multi_hotspots_ref.png"), dpi=300, bbox_inches='tight', transparent=True)


    data = pd.DataFrame(data=res).to_numpy()
    mean_jaccard = data.mean(axis=1)
    cm = sns.color_palette("Reds", as_cmap=True)
    colors = [cm(x) for x in mean_jaccard]

    fig, ax = plt.subplots(1, 1, figsize=[2, 2])

    nest.plot.multi_hotspots(adata_full, color_type=colors, ax=ax, show=False, show_colorbar=True)
    fig.savefig(os.path.join(image_save_dir, f"subsampling_{frac_genes}.pdf"), dpi=300, transparent=True, bbox_inches="tight")
    plt.show()

